#! /usr/bin/env python

# TuxumCluster python library 
# License: GPL v2
# Author: Tuxum Secure Systems, S. L.

import os
import re
import socket
import os.path
import tuxumssh
import string
import getpass

#Default values
DRBDCONF="/etc/drbd.conf"
DRBDSCRIPT="/etc/init.d/drbd"
DRBDADM="drbdadm"
DRBDSETUP="drbdsetup"
HEARTBEATSCRIPT="/etc/init.d/heartbeat"
HA_CF="/etc/ha.d/ha.cf"
AUTHKEYS="/etc/ha.d/authkeys"
CLUSTERCONF="/etc/ocfs2/cluster.conf"
GFS2CLUSTERCONF="/etc/cluster/cluster.conf"
TUXUMCLUSTERDIR="/etc/tuxum-cluster"
TUXUMCLUSTERSCRIPT="/etc/init.d/tuxum-cluster"
THISNODECONF="/etc/tuxum-cluster/thisnode.conf"
OTHERNODECONF="/etc/tuxum-cluster/othernode.conf"
OCFS2DIR="/etc/ocfs2"
OCFS2SCRIPT="/etc/init.d/o2cb"
MKFSOCFS2="mkfs.ocfs2"
GFS2DIR="/etc/cluster"
GFS2SCRIPT="/etc/init.d/cman"
MKFSGFS2="mkfs.gfs2"
HOSTS="/etc/hosts"
SED="/bin/sed"
FSTAB="/etc/fstab"

DEFAULT_RESOURCEPORT=7789
DEFAULT_OCFS2PORT=7777
DEFAULT_SSH_TIMEOUT=60

# Error definitions
OK=0
NO_ERROR=0
ERROR_UNSPECIFIED=1
ERROR_RESOURCE_EXISTS=2
ERROR_RESOURCEPORT_EXISTS=3
ERROR_DRBDDEVICE_EXISTS=4
ERROR_INVALID_NODECONF_FILE=5
ERROR_FIRST_STORAGE_DEVICE_EXISTS=6
ERROR_SECOND_STORAGE_DEVICE_EXISTS=7
ERROR_MOUNTPOINT_NOT_EMPTY=8
ERROR_INVALID_MOUNTPOINT=9

class DrbdDevice:
	"""
	DRBD 8 device definition class
	"""
	def __init__(self, disk, address, port):
		"""
		DRBD 8 Device definition
		"""
		self.disk = disk
		self.address = address
		self.port = port



class ClusterNode:
	"""
	OCFS2 and Heartbeat node definition class
	"""
	def __init__(self, name, address, port, interface):
		"""
		DRBD 8 Device definition
		"""
		self.name = name
		self.address = address
		self.port = port
		self.interface = interface



class AdmCluster:
	"""
	Cluster administration tool main class
	"""

	def initDrbd8(self):
		"""
		Initializes DRBD 8 cluster
		"""
		if os.path.exists(DRBDCONF):
			os.system( "mv " + DRBDCONF + " " + DRBDCONF + ".old" )
		drbdconfFile = open( DRBDCONF, 'w' )
		cadena = \
"""global
{
	usage-count no;
}

common
{
	syncer
	{
		rate 20M;
	}
}
"""
		drbdconfFile.write(cadena)
		drbdconfFile.close()



	def writeNodeConfFiles(self, thisNodeAddress, otherNodeAddress, otherNodeName):
		"""
		Writes node configuration files
		"""
		os.system( "mkdir -p " + TUXUMCLUSTERDIR )
		nodesconf = open( OTHERNODECONF, "w")
		nodesconf.write(otherNodeAddress + "\n")
		nodesconf.write(otherNodeName + "\n")
		nodesconf.close()
		nodesconf = open( THISNODECONF, "w")
		nodesconf.write(thisNodeAddress + "\n")
		nodesconf.write(socket.gethostname() + "\n")
		nodesconf.close()



	def addDrbd8(self, resource, deviceName, theOtherHostName, thisDevice, theOtherDevice):
		"""
		Adds a DRBD 8 device
		"""
		cadena = \
"""
resource %s
{
	protocol C;
	net
	{
		allow-two-primaries;
		after-sb-0pri discard-zero-changes;
		after-sb-1pri discard-secondary;
		after-sb-2pri disconnect;
	}

	startup
	{ 
		become-primary-on both;
	}

	handlers
	{
#		split-brain "/usr/lib/drbd/notify-split-brain.sh root";
	}

	on %s
	{
		device		%s;
		disk		%s;
		address		%s:%s;
		meta-disk	internal;
	}

	on %s
	{
		device		%s;
		disk		%s;
		address		%s:%s;
		meta-disk	internal;
	}
}
"""%(resource, socket.gethostname(), deviceName, thisDevice.disk, thisDevice.address, thisDevice.port, theOtherHostName, deviceName, theOtherDevice.disk, theOtherDevice.address, theOtherDevice.port )
		drbdconfFile = open( DRBDCONF, 'a' )
		drbdconfFile.write( cadena )
		drbdconfFile.close()

		"""
		Try to create DRBD device
		"""
		os.system( DRBDADM + " create-md " + resource )
		os.system( DRBDADM + " attach " + resource )


	def addDrbdToModules(self):
		"""
		Load DRBD module and put it into /etc/modules
		"""
		os.system("modprobe drbd")
		os.system('cat /etc/modules | grep -w -v "drbd" > /tmp/etcmodules.tmp')
		os.system("echo drbd >> /tmp/etcmodules.tmp")
		os.system("mv /tmp/etcmodules.tmp /etc/modules")



	def addToHosts(self, thisAddress, thisName, otherAddress, otherName):
		"""
		Put IPs and names on /etc/hosts
		"""
		exp = re.compile(r"" + thisAddress + "\s" + thisName, re.IGNORECASE)
		if not self.searchForString( HOSTS, exp ):
			os.system( 'echo ' + thisAddress + '\\\t' + thisName + ' >> ' + HOSTS )
		exp = re.compile(r"" + otherAddress + "\s" + otherName, re.IGNORECASE)
		if not self.searchForString( HOSTS, exp ):
			os.system( 'echo ' + otherAddress + '\\\t' + otherName + ' >> ' + HOSTS )



	def synchronize(self):
		"""
		Synchronizes defined devices for added resources, deleting all data on peer
		"""
		os.system( DRBDADM + " -- --overwrite-data-of-peer primary all")



	def configureHeartbeat(self, thisHostName, theOtherHostName, interface):
		"""
		Configures Heartbeat
		"""
		hacf = open( HA_CF, 'w' )
		hacf.write( "node\t" + thisHostName + " " + theOtherHostName + "\n" )
		hacf.write( "ucast\t" + interface + " " + theOtherHostName + "\n" )
		hacf.write( "crm\trespawn\n" )
		hacf.close()

		authkeys = open( AUTHKEYS, 'w' )
		authkeys.write( "auth 1\n" )
		authkeys.write( "1 crc\n" )
		authkeys.close()
		os.system( "chmod 0600 " + AUTHKEYS )
		os.system( "chown root:root " + AUTHKEYS )



	def restartHeartbeat(self):
		"""
		Restarts Heartbeat
		"""
		os.system( HEARTBEATSCRIPT + " restart" )



	def configureGfs2(self, thisDevice, theOtherDevice):
		"""
		Configures a system for building an GFS2 filesystem
		"""
		os.system( "mkdir -p " + GFS2DIR )

		cadena = \
"""<?xml version="1.0"?>
<cluster name="tuxumcluster" config_version="1">
  <!-- post_join_delay: number of seconds the daemon will wait before
                        fencing any victims after a node joins the domain
       post_fail_delay: number of seconds the daemon will wait before
            	        fencing any victims after a domain member fails
       clean_start    : prevent any startup fencing the daemon might do.
		        It indicates that the daemon should assume all nodes
		        are in a clean state to start. -->
  <fence_daemon clean_start="0" post_fail_delay="0" post_join_delay="3"/>
  <clusternodes>
    <clusternode name="%s" votes="1" nodeid="1">
      <fence>
        <!-- Handle fencing manually -->
        <method name="human">
          <device name="human" nodename="%s"/>
        </method>
      </fence>
    </clusternode>
    <clusternode name="%s" votes="1" nodeid="2"> 
      <fence>
        <!-- Handle fencing manually -->
        <method name="human">
          <device name="human" nodename="%s"/>
        </method>
      </fence>
    </clusternode>
  </clusternodes>
  <!-- cman two nodes specification -->
  <cman expected_votes="1" two_node="1"/>
  <fencedevices>
    <!-- Define manual fencing -->
    <fencedevice name="human" agent="fence_manual"/>
  </fencedevices>
</cluster>
"""%(thisDevice.name, thisDevice.name, theOtherDevice.name, theOtherDevice.name)
		clusterconf = open( GFS2CLUSTERCONF, "w" )
		clusterconf.write( cadena )
		clusterconf.close()



	def configureOcfs2(self, thisDevice, theOtherDevice):
		"""
		Configures a system for building an OCFS2 filesystem
		"""
		os.system( "mkdir -p " + OCFS2DIR )
		thisOcfs2Port = thisDevice.port
		theOtherOcfs2Port = theOtherDevice.port

		if not thisOcfs2Port:
			thisOcfs2Port = DEFAULT_OCFS2PORT

		if not theOtherOcfs2Port:
			theOtherOcfs2Port = DEFAULT_OCFS2PORT

		cadena = \
"""
node:
	ip_port = %s
	ip_address = %s
	number = 0
	name = %s
	cluster = ocfs2

node:
	ip_port = %s
	ip_address = %s
	number = 1
	name = %s
	cluster = ocfs2

cluster:
	node_count = 2
	name = ocfs2
"""%(thisOcfs2Port, thisDevice.address, thisDevice.name, theOtherOcfs2Port, theOtherDevice.address, theOtherDevice.name)
		clusterconf = open( CLUSTERCONF, "w" )
		clusterconf.write( cadena )
		clusterconf.close()



	def formatOcfs2(self, device):
		"""
		Formats a device with an OCFS2 filesystem
		"""
		os.system( MKFSOCFS2 + " " + device )



	def formatGfs2(self, device, resourceName):
		"""
		Formats a device with a GFS2 filesystem
		"""
		os.system( MKFSGFS2 + " -t tuxumcluster:" + resourceName + " -p lock_dlm -j 2 " + device )



	def restartOcfs2(self):
		"""
		Restarts OCFS2 filesystems
		"""

		os.system( OCFS2SCRIPT + " restart" )
		os.system( OCFS2SCRIPT + " online" )



	def restartGfs2(self):
		"""
		Restarts GFS2 filesystems
		"""

		os.system( GFS2SCRIPT + " start" )



	def searchForString(self, fileName, expression):
		"""
		Searches a string (not case sensitive) on a file
		"""

		if os.path.exists(fileName):
			file = open( fileName, 'r' )
			fileContent = file.read()
			file.close()

			return (re.search(expression, fileContent))
		else:
			return False


	def removeResource( self, resourceToRemove ):
		"""
		Removes a resource from the cluster
		"""

		thisAddress, thisName, otherAddress, otherName = self.readNodeConfFiles()
		ssh = tuxumssh.TuxumSSH( thisAddress, otherAddress )

		os.system( "umount `drbdadm sh-dev " + resourceToRemove + "`" )
		ssh.sendline( "umount `drbdadm sh-dev " + resourceToRemove + "`" )

		os.system ( DRBDADM + " down " + resourceToRemove )
		ssh.sendline( DRBDADM + " down " + resourceToRemove )

		sedString = "s/resource " + resourceToRemove + "\\b/skip " + resourceToRemove + "/g"
		os.system( SED + ' -i -e "' + sedString + '" ' + DRBDCONF )
		os.system( "scp " + DRBDCONF + " root@" + otherAddress + ":" + DRBDCONF )




	def configureFirstNode( self, firstNodeAddress, secondNodeName, secondNodeAddress, interface, ocfs2port ):
		"""
		Configures the first node of the cluster
		firstNodeAddress and secondNodeAddress are the IP addresses of the first and second node respectively.
		secondNodeName is the host name of the second node
		interface is the name of the network interface on this host (eth1, for example)
		ocfs2port is the port address that OCFS2 will use
		"""
		self.addToHosts(firstNodeAddress, socket.gethostname(), secondNodeAddress, secondNodeName)
		self.initDrbd8()
		self.configureHeartbeat(socket.gethostname(), secondNodeName, interface)

		thisClusterNode = ClusterNode(socket.gethostname(), firstNodeAddress, ocfs2port, interface)
		otherClusterNode = ClusterNode(secondNodeName, secondNodeAddress, ocfs2port, interface)

		self.configureOcfs2(thisClusterNode, otherClusterNode)
		self.configureGfs2(thisClusterNode, otherClusterNode)
		self.writeNodeConfFiles( firstNodeAddress, secondNodeAddress, secondNodeName )



	def addResource( self, firstStorageDevice, mountPoint, secondStorageDevice, fileSystem, _resourceName, _drbdDevice, _resourcePort ):
		"""
		Adds a resource to the configured first node of a cluster
		firstStorageDevice and secondStorageDevice are the devices that will conform the resource on both nodes
		resourceName is the (unique) name of the resource
		drbdDevice is the device name of the DRBD device
		resource port is the DRBD port for this resource	
		"""
		if os.path.exists(mountPoint):
			if not os.path.isdir(mountPoint):
				return ERROR_INVALID_MOUNTPOINT
		else:
			os.system("mkdir " + mountPoint)
			if not os.path.isdir(mountPoint):
				return ERROR_INVALID_MOUNTPOINT

		dirList = os.listdir(mountPoint)

		if len(dirList) > 0:
			return ERROR_MOUNTPOINT_NOT_EMPTY



		exp = re.compile(r"disk\s+" + firstStorageDevice + ";", re.IGNORECASE)
		if self.searchForString( DRBDCONF, exp):
			return ERROR_FIRST_STORAGE_DEVICE_EXISTS

		exp = re.compile(r"disk\s+" + secondStorageDevice + ";", re.IGNORECASE)
		if self.searchForString( DRBDCONF, exp):
			return ERROR_SECOND_STORAGE_DEVICE_EXISTS
		resourceName = _resourceName
		if not resourceName:
			n = 0
			exp = re.compile(r"resource\s+r" + str(n), re.IGNORECASE)
			while self.searchForString( DRBDCONF, exp):
				n = n + 1
				exp = re.compile(r"resource\s+r" + str(n), re.IGNORECASE)
			resourceName = "r" + str(n)
		else:
			exp = re.compile(r"resource\s+" + resourceName, re.IGNORECASE)
			if self.searchForString( DRBDCONF, exp):
				return ERROR_RESOURCE_EXISTS

		resourcePort = _resourcePort
		if not resourcePort:
			n = DEFAULT_RESOURCEPORT
			exp = re.compile(r"address\s+\d+.\d+.\d+.\d+:" + str(n) + ";", re.IGNORECASE)
			while self.searchForString( DRBDCONF, exp):
				n = n + 1
				exp = re.compile(r"address\s+\d+.\d+.\d+.\d+:" + str(n) + ";", re.IGNORECASE)
			resourcePort = n
		else:
			exp = re.compile(r"address\s+\d+.\d+.\d+.\d+:" + resourcePort + ";", re.IGNORECASE)
			if self.searchForString( DRBDCONF, exp):
				return ERROR_RESOURCEPORT_EXISTS

		drbdDevice = _drbdDevice
		if not drbdDevice:
			n = 0
			exp = re.compile(r"device\s+/dev/drbd" + str(n) + ";", re.IGNORECASE)
			while self.searchForString( DRBDCONF, exp):
				n = n + 1
				exp = re.compile(r"device\s+/dev/drbd" + str(n) + ";", re.IGNORECASE)
			drbdDevice = "/dev/drbd" + str(n)
		else:
			exp = re.compile(r"device\s+" + drbdDevice + ";", re.IGNORECASE)
			if self.searchForString( DRBDCONF, exp):
				return ERROR_DRBDDEVICE_EXISTS

		fstabString = drbdDevice + "\t" + mountPoint + "\t" + fileSystem + "\t_netdev,defaults\t0\t0\n"
		fstab = open( FSTAB, "a" )
		fstab.write( fstabString )
		fstab.close()

		thisAddress, thisName, otherAddress, otherName = self.readNodeConfFiles()

		if (thisAddress, thisName, otherAddress, otherName) == (None, None, None, None):
			return ERROR_INVALID_NODECONF_FILE
		else:
			thisDevice = DrbdDevice(firstStorageDevice, thisAddress, resourcePort)
			otherDevice = DrbdDevice(secondStorageDevice, otherAddress, resourcePort)

			self.addDrbdToModules()
			self.addDrbd8(resourceName, drbdDevice, otherName, thisDevice, otherDevice)
			self.synchronize()

			if (fileSystem == "ocfs2"):
				self.restartHeartbeat()
				self.restartOcfs2()
				self.formatOcfs2(drbdDevice)

			if (fileSystem == "gfs2"):
				self.restartGfs2()
				self.formatGfs2(drbdDevice, resourceName)

			os.system( DRBDADM + " connect " + resourceName )
			os.system( "sleep 1")
			os.system ( DRBDADM + " primary " + resourceName )
			os.system( "mount -a -t ocfs2" )
			os.system( "mount -a -t gfs2" )

			return NO_ERROR



	def readNodeConfFiles( self ):
		"""
		Reads Tuxum node conf files
		"""
		if os.path.exists(THISNODECONF) and os.path.exists(OTHERNODECONF):
			thisNode = open( THISNODECONF, "r" )
			thisNodeList = thisNode.readlines()
			thisNode.close()
			otherNode = open( OTHERNODECONF, "r" )
			otherNodeList = otherNode.readlines()
			otherNode.close()
			if len(thisNodeList) < 2 or len(otherNodeList) < 2:
				return (None, None, None, None)
			else:
				thisAddress = thisNodeList[0]
				thisName = thisNodeList[1]
				otherAddress = otherNodeList[0]
				otherName = otherNodeList[1]
				return (thisAddress.strip(), thisName.strip(), otherAddress.strip(), otherName.strip())
		else:
			return (None, None, None, None)



	def configureSecondNode( self, firstNodeIP, interface ):
		"""
		Configures a second cluster node using SSH
		"""
		os.system( "mkdir -p " + OCFS2DIR )
		os.system( "mkdir -p " + TUXUMCLUSTERDIR )

		ssh = tuxumssh.TuxumSSH( "localhost", firstNodeIP )
		ssh.sshOrgy()

		os.system( TUXUMCLUSTERSCRIPT + " stop" )
		if os.path.exists( "/proc/drbd" ):
			os.system( DRBDADM + " down all" )
			os.system( DRBDSCRIPT + " stop" )
			
		for i in DRBDCONF, AUTHKEYS, OTHERNODECONF, THISNODECONF, CLUSTERCONF, GFS2CLUSTERCONF:
			confFile = ssh.sshGet(i)
			self.writeFile(confFile, i)

		fstabFile = ssh.sshGet(FSTAB)
		self.writeFile( fstabFile, "/etc/fstab.tmp" )

		os.system( "sync")
		os.system( "cat /etc/fstab.tmp | grep '/dev/drbd' > /etc/fstab.tmp2" )
		os.system( "cat " + FSTAB + " | grep -v '/dev/drbd' > /etc/fstab.tmp3" )
		os.system( "cat /etc/fstab.tmp3 /etc/fstab.tmp2 > " + FSTAB )
		os.system( "rm /etc/fstab.tmp /etc/fstab.tmp2 /etc/fstab.tmp3" )
		os.system( "for i in `cat /etc/fstab | grep '/dev/drbd' | awk '{print $2}'`; do mkdir -p $i; done" )
		
		os.system( "mv " + OTHERNODECONF + " " + THISNODECONF + ".tmp")
		os.system( "mv " + THISNODECONF + " " + OTHERNODECONF )
		os.system( "mv " + THISNODECONF + ".tmp " + THISNODECONF )
		os.system( "chmod 0600 " + AUTHKEYS )
		os.system( "chown root:root " + AUTHKEYS )
			
		thisAddress, thisName, otherAddress, otherName = self.readNodeConfFiles()
		if (thisAddress, thisName, otherAddress, otherName) == (None, None, None, None):
			print "Error: Invalid node configuration file (check " + THISNODECONF + " and " + OTHERNODECONF + ")"
		else:
			hacf = open( HA_CF, 'w' )
			hacf.write( "node\t" + thisName + " " + otherName + "\n" )
			hacf.write( "ucast\t" + interface + " " + otherName + "\n" )
			hacf.write( "crm\trespawn\n" )
			hacf.close()
				
		self.addDrbdToModules()				

		self.addToHosts( thisAddress, thisName, otherAddress, otherName )

		ssh.localAddress = thisAddress

		if os.path.exists( "/proc/drbd" ):
			os.system( DRBDADM + " down all" )
			os.system( "modprobe -r drbd" )

		os.system( "modprobe drbd" )
		os.system( DRBDADM + " --force wipe-md all" )
		os.system( DRBDADM + " create-md all" )
		os.system( DRBDADM + " attach all" )
		os.system( DRBDADM + " secondary all" )
		os.system( DRBDADM + " -- --discard-my-data connect all" )

		ssh.sshOrgy( thisName, otherName )
		ssh.sendline ( DRBDADM + " disconnect all" )
		ssh.sendline ( DRBDADM + " -- --overwrite-data-of-peer primary all" )
		ssh.sendline ( DRBDADM + " connect all" )

		os.system( "sleep 5")

		os.system ( DRBDADM + " primary all" )

		os.system( TUXUMCLUSTERSCRIPT + " start" )

		os.system( DRBDADM + " adjust all" )

	def recoverSplitBrain( self ):
		"""
		Recovers split brain discarding this node
		"""

		thisAddress, thisName, otherAddress, otherName = self.readNodeConfFiles()

		ssh = tuxumssh.TuxumSSH( "localhost", otherAddress )

		ssh.localAddress = thisAddress

		if os.path.exists( "/proc/drbd" ):
			os.system( DRBDADM + " down all" )
			os.system( "modprobe -r drbd" )

		os.system( "modprobe drbd" )
		os.system( DRBDADM + " attach all" )
		os.system( DRBDADM + " secondary all" )
		os.system( DRBDADM + " -- --discard-my-data connect all" )

		ssh.sshOrgy( thisName, otherName )
		ssh.sendline ( DRBDADM + " disconnect all" )
		ssh.sendline ( DRBDADM + " -- --overwrite-data-of-peer primary all" )
		ssh.sendline ( DRBDADM + " connect all" )

		os.system( "sleep 5")

		os.system ( DRBDADM + " primary all" )

		os.system( TUXUMCLUSTERSCRIPT + " start" )

	def writeFile(self, content, file):
		file = open( file, "w" )
                file.write( content )
                file.close()
